import einops
from kappamodules.layers import ContinuousSincosEmbed, LinearProjection
from kappamodules.transformer import PerceiverBlock
from torch import nn


class PerceiverDecoder(nn.Module):
    def __init__(
            self,
            dim,
            num_heads,
            ndim,
            input_dim,
            output_dim,
            init_weights="truncnormal002",
            eps=1e-6,
    ):
        super().__init__()
        # create query
        self.pos_embed = ContinuousSincosEmbed(dim=dim, ndim=ndim)
        self.query = nn.Sequential(
            LinearProjection(dim, dim, init_weights=init_weights),
            nn.GELU(),
            LinearProjection(dim, dim, init_weights=init_weights),
        )
        # perceiver
        self.proj = LinearProjection(input_dim, dim, init_weights=init_weights, optional=True)
        self.perc = PerceiverBlock(dim=dim, num_heads=num_heads, init_weights=init_weights, eps=eps)
        self.norm = nn.LayerNorm(dim, eps=eps)
        self.pred = LinearProjection(dim, output_dim, init_weights=init_weights)

    def forward(self, x, pos, block_kwargs=None, unbatch_mask=None):
        if pos is None:
            assert not self.training, f"{type(self).__name__} expects query positions during training"
            return None
        assert x.ndim == 3
        assert pos.ndim == 3

        # create query
        query = self.query(self.pos_embed(pos))

        # project to perceiver dim
        x = self.proj(x)

        # perceiver
        x = self.perc(q=query, kv=x, **(block_kwargs or {}))

        # predict value
        x = self.norm(x)
        x = self.pred(x)

        # dense tensor (batch_size, max_num_points, dim) -> sparse tensor (batch_size * num_points, dim)
        x = einops.rearrange(x, "batch_size max_num_points dim -> (batch_size max_num_points) dim")
        if len(pos) == 1:
            # batch_size=1 -> no padding is needed
            pass
        else:
            if unbatch_mask is not None:
                x = x[unbatch_mask]
        return x



from torch import nn

from kappamodules.init import init_xavier_uniform_merged_linear, init_truncnormal_zero_bias
from kappamodules.utils.shapes import to_ndim


class Film(nn.Module):
    def __init__(self, dim_cond, dim_out, channel_first=True, init_weights="xavier_uniform"):
        super().__init__()
        self.dim_cond = dim_cond
        self.dim_out = dim_out
        self.channel_first = channel_first
        self.modulation = nn.Linear(dim_cond, dim_out * 2)
        self.init_weights = init_weights
        self.reset_parameters()

    def reset_parameters(self):
        if self.init_weights == "torch":
            pass
        elif self.init_weights == "xavier_uniform":
            init_xavier_uniform_merged_linear(self.modulation, num_layers=2)
        elif self.init_weights in ["truncnormal", "truncnormal002"]:
            self.apply(init_truncnormal_zero_bias)
        else:
            raise NotImplementedError

    def forward(self, x, cond):
        mod = self.modulation(cond)
        if self.channel_first:
            scale, shift = to_ndim(mod, ndim=x.ndim).chunk(2, dim=1)
        else:
            scale, shift = to_ndim(mod, ndim=x.ndim, pad_dim=1).chunk(2, dim=-1)
        return x * (scale + 1) + shift



from torch import nn


class Sequential(nn.Sequential):
    """
    torch.nn.Sequential but one can pass arbitrary arguments
    net = nn.Sequential(...)
    net(a, b) -> this fails

    net = km.Sequential(...)
    net(a, b) -> passes a and b to all layers in km.Sequential
    """

    def forward(self, x, *args, **kwargs):
        for module in self:
            x = module(x, *args, **kwargs)
        return x


import torch.nn.functional as F
from torch import nn

from kappamodules.init import init_xavier_uniform_zero_bias, init_truncnormal_zero_bias
from kappamodules.utils.param_checking import to_2tuple


class SwigluMlp(nn.Module):
    def __init__(
            self,
            in_dim,
            hidden_dim=None,
            out_dim=None,
            bias=True,
            init_weights="xavier_uniform",
            init_last_proj_zero=False,
            match_mlp_params=True,
    ):
        super().__init__()
        self.init_weights = init_weights
        self.init_last_proj_zero = init_last_proj_zero
        out_dim = out_dim or in_dim
        hidden_dim = hidden_dim or in_dim
        if match_mlp_params:
            # reduce hidden dim to make number of parameters equivalent to a non swiglu MLP
            hidden_dim = (int(hidden_dim * 2 / 3) + 7) // 8 * 8
        bias1, bias2 = to_2tuple(bias)

        self.fc1 = nn.Linear(in_dim, 2 * hidden_dim, bias=bias1)
        self.fc2 = nn.Linear(hidden_dim, out_dim, bias=bias2)
        self.reset_parameters()

    def reset_parameters(self):
        if self.init_weights == "torch":
            pass
        elif self.init_weights == "xavier_uniform":
            self.apply(init_xavier_uniform_zero_bias)
        elif self.init_weights in ["truncnormal", "truncnormal002"]:
            self.apply(init_truncnormal_zero_bias)
        else:
            raise NotImplementedError
        if self.init_last_proj_zero:
            nn.init.zeros_(self.fc2.weight)

    def forward(self, x):
        x1, x2 = self.fc1(x).chunk(2, dim=-1)
        x = F.silu(x1) * x2
        x = self.fc2(x)
        return x
